# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    
#    Copyright (c) 2013 Noviat nv/sa (www.noviat.be). All rights reserved.
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv, fields
from tools.translate import _
import logging
_logger = logging.getLogger(__name__)

class account_move(osv.osv):
    _inherit = 'account.move'

    def button_cancel(self, cr, uid, ids, context=None):
        for move in self.browse(cr, uid, ids, context=context):
            for move_line in move.line_id:
                st = move_line.statement_id
                if st and st.state == 'confirm':
                    raise osv.except_osv('Warning', _('Operation not allowed ! \
                        \nYou cannot unpost an Accounting Entry that is linked to a Confirmed Bank Statement.'))
        return super(account_move, self).button_cancel(cr, uid, ids, context=context)

    def button_dummy(self, cr, uid, ids, context=None):
        return self.write(cr, uid, ids, {}, context=context)
    
account_move()

class account_move_line(osv.osv):
    _inherit = 'account.move.line'
    
    def onchange_tax_code(self, cr, uid, ids, tax_code_id, debit, credit):
        val = {}
        if tax_code_id:
            if debit or credit:
                val['tax_amount'] = debit or credit
            else:
                 move_lines = self.browse(cr, uid, ids)
                 for move_line in move_lines:
                     if move_line.debit or move_line.credit:
                        val['tax_amount'] = move_line.debit or move_line.credit
        else:
            val['tax_amount'] = False
        return {'value': val}

    def fields_view_get(self, cr, uid, view_id=None, view_type='form', context=None, toolbar=False, submenu=False):
        #_logger.warn('fields_view_get, context = %s, view_type= %s', context, view_type)
        mod_obj = self.pool.get('ir.model.data')
        if context is None:
            context = {}
        if context.get('act_window_from_bank_statement', False):
            if view_type=='tree':
                model_data_ids = mod_obj.search(cr, uid,[('model', '=', 'ir.ui.view'), ('module', '=', 'account_bank_statement_voucher'), ('name', '=', 'view_move_line_reconcile_tree')], context=context)
                view_id = mod_obj.read(cr, uid, model_data_ids, fields=['res_id'], context=context)[0]['res_id']        
                context.update({'view_mode':'tree'})
            elif view_type=='search':
                model_data_ids = mod_obj.search(cr, uid,[('model', '=', 'ir.ui.view'), ('module', '=', 'account_bank_statement_voucher'), ('name', '=', 'view_move_line_reconcile_search')], context=context)
                view_id = mod_obj.read(cr, uid, model_data_ids, fields=['res_id'], context=context)[0]['res_id']        
                context.update({'view_mode':'search'})
        return super(account_move_line, self).fields_view_get(cr, uid, view_id, view_type, context=context, toolbar=toolbar, submenu=submenu)

    def search(self, cr, uid, args, offset=0, limit=None, order=None, context=None, count=False):
        if context is None:
            context = {}
        if args and context.get('act_window_from_bank_statement'):
            context_ids = []
            if context.get('entries'):
                context_ids = [x['id'] for x in context['entries']]
            #_logger.warn('context_ids=%s', context_ids)
            search_ids = []
            query2 = ''
            # pass one : compute intersection of ids in args and context
            new_args1 = []
            for arg in args:
                if type(arg) in [list,tuple] and len(arg) == 3:                   
                    if arg[0] == 'id':
                        if arg[1] == 'in':
                            search_ids = arg[2]
                        elif arg[1] == '=':
                            search_ids = [arg[2]]
                    else:
                        new_args1.append(arg)
            in_ids = [x for x in context_ids if x in search_ids]
            # pass 2 : refine query with invoice number matching
            new_args2 = []
            id_flag = False 
            for arg in new_args1:
                if type(arg) in [list,tuple] and len(arg) == 3:  
                    if arg[0] == 'invoice' and arg[2]:
                        if in_ids:
                            query2 = "AND aml.id IN (%s)" % ','.join(['%s'] * len(in_ids)) %tuple(in_ids)
                        query = "SELECT aml.id " \
                           "FROM account_move_line aml " \
                           "INNER JOIN account_invoice ai ON aml.move_id = ai.move_id " \
                           "WHERE ai.number ILIKE '%" + arg[2] + "%'"
                        #_logger.warn(query)
                        cr.execute(query + query2)
                        res = cr.fetchall()
                        out_ids = [x[0] for x in res]
                        new_args2.append(('id', 'in', out_ids))
                        id_flag = True
                    else:
                        new_args2.append(arg)                
            #_logger.warn('search exit with new args=%s', new_args2)
            if not id_flag and in_ids:
                new_args2.append(('id', 'in', in_ids))
            return super(account_move_line, self).search(cr, uid, new_args2, offset, limit, order, context, count)
            
        res = super(account_move_line, self).search(cr, uid, args, offset, limit, order, context, count)
        #_logger.warn('search exit with old args=%s', args)
        return res

    def unlink(self, cr, uid, ids, context=None, check=True):
        for move_line in self.browse(cr, uid, ids, context):
            st = move_line.statement_id
            if st and st.state == 'confirm':
                raise osv.except_osv('Warning', _('Operation not allowed ! \
                    \nYou cannot delete an Accounting Entry that is linked to a Confirmed Bank Statement.'))
        return super(account_move_line, self).unlink(cr, uid, ids, context=context, check=check)

    def write(self, cr, uid, ids, vals, context=None, check=True, update_check=True):
        #_logger.warn('write, vals = %s', vals)
        for move_line in self.browse(cr, uid, ids, context):
            st = move_line.statement_id
            if st and st.state == 'confirm':
                if vals.keys() not in [['reconcile_id'],['reconcile_partial_id']]:
                    raise osv.except_osv('Warning', _('Operation not allowed ! \
                        \nYou cannot modify an Accounting Entry that is linked to a Confirmed Bank Statement.'))
        return super(account_move_line, self).write(cr, uid, ids, vals, context, check, update_check)

account_move_line()