# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    
#    Copyright (c) 2012 Noviat nv/sa (www.noviat.be). All rights reserved.
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
from tools.translate import _
from osv import fields, osv
import logging
_logger = logging.getLogger(__name__)

class nov_print_journal_fiscalyear(osv.osv_memory):
    _name = 'nov.print.journal.fiscalyear' 
    _description = 'l10n_be_coa_multilang - Print Journal by Fiscal Year'

    def _get_fiscalyear(self, cr, uid, context=None):
        now = time.strftime('%Y-%m-%d')
        fiscalyears = self.pool.get('account.fiscalyear').search(cr, uid, [('date_start', '<', now), ('date_stop', '>', now)], limit=1 )
        return fiscalyears and fiscalyears[0] or False

    _columns = {
        'fiscalyear_id': fields.many2one('account.fiscalyear', 'Fiscal Year', required=True),
        'journal_id': fields.many2one('account.journal', 'Journal', required=True),
        'sort': fields.selection([('date','Date'),('number','Number')],
            'Entries Sorted By', required=True),
        }
    
    _defaults={
        'fiscalyear_id': _get_fiscalyear,
        'sort': lambda *a: 'number'
            }

    def print_report(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        obj_period = self.pool.get('account.period')
        obj_move = self.pool.get('account.move')
        data = self.read(cr, uid, ids)[0]       
        fiscalyear_id = data['fiscalyear_id'][0]
        journal_id = data['journal_id'][0]
        sort = data['sort']       
        period_ids = obj_period.search(cr,uid,[('fiscalyear_id', '=', fiscalyear_id)])
        aml_ids = obj_move.search(cr, uid, [('journal_id', '=', journal_id), ('period_id', 'in', period_ids)], limit=1)
        if not aml_ids:
            raise osv.except_osv(_('No Data Available'), _('No records found for your selection!'))   
        datas = {'ids': []}
        datas['model'] = 'account.fiscalyear'
        datas['fiscalyear_id'] = fiscalyear_id
        datas['journal_id'] = journal_id
        datas['sort'] = sort
        
        return {
            'type': 'ir.actions.report.xml',
            'report_name': 'nov.account.journal.fiscalyear.print',
            'datas': datas,
        }

nov_print_journal_fiscalyear()

