# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    
#    Copyright (c) 2012 Noviat nv/sa (www.noviat.be). All rights reserved.
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
from tools.translate import _
from osv import fields, osv
import logging
_logger = logging.getLogger(__name__)

class wiz_partner_open_arap_fy(osv.osv_memory):
    _name = 'wiz.partner.open.arap.fy' 
    _description = 'Print Open Receivables/Payables by Fiscal Year'

    def _get_fiscalyear(self, cr, uid, context=None):
        if context.get('active_model') == 'account.fiscalyear' and context.get('active_ids'):
            if len(context['active_ids']) == 1:
               return context['active_ids'][0]
        now = time.strftime('%Y-%m-%d')
        fiscalyears = self.pool.get('account.fiscalyear').search(cr, uid, [('date_start', '<', now), ('date_stop', '>', now)], limit=1 )
        return fiscalyears and fiscalyears[0] or False

    _columns = {
        'fiscalyear_id': fields.many2one('account.fiscalyear', 'Fiscal year', required=True),
        'target_move': fields.selection([('posted', 'All Posted Entries'),
                                         ('all', 'All Entries'),
                                        ], 'Target Moves', required=True),
        'result_selection': fields.selection([('customer','Receivable Accounts'),
                                              ('supplier','Payable Accounts'),
                                              ('customer_supplier','Receivable and Payable Accounts')],
                                              "Partner's", required=True),
        'company_id': fields.many2one('res.company', 'Company', readonly=True)
    }    
    _defaults={
        'fiscalyear_id': _get_fiscalyear,
        'target_move': 'posted',
        'result_selection': 'customer',
        'company_id': lambda s,cr,uid,c: s.pool.get('res.company')._company_default_get(cr, uid, 'account.coda', context=c),
    }

    def print_report(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        data = self.read(cr, uid, ids)[0]
        fiscalyear_id = data['fiscalyear_id'][0]
        company_id = data['company_id'][0]        
        data.update( {
            'ids': [fiscalyear_id], 
            'fiscalyear_id': fiscalyear_id, 
            'company_id': company_id,        
            'model': 'account.fiscalyear'
        } )
        return {
            'type': 'ir.actions.report.xml',
            'report_name': 'account.partner.open.arap.fy.print',
            'datas': data,
        }

wiz_partner_open_arap_fy()

class wiz_partner_open_arap_period(osv.osv_memory):
    _name = 'wiz.partner.open.arap.period' 
    _description = 'Print Open Receivables/Payables by Period'

    def _get_period(self, cr, uid, context=None):
        if context.get('active_model') == 'account.period' and context.get('active_ids'):
            if len(context['active_ids']) == 1:
               return context['active_ids'][0]
        now = time.strftime('%Y-%m-%d')
        periods = self.pool.get('account.period').search(cr, uid, [('date_start', '<', now), ('date_stop', '>', now)], limit=1 )
        return periods and periods[0] or False
    
    _columns = {
        'period_id': fields.many2one('account.period', 'Period', required=True),
        'target_move': fields.selection([('posted', 'All Posted Entries'),
                                         ('all', 'All Entries'),
                                        ], 'Target Moves', required=True),
        'result_selection': fields.selection([('customer','Receivable Accounts'),
                                              ('supplier','Payable Accounts'),
                                              ('customer_supplier','Receivable and Payable Accounts')],
                                              "Partner's", required=True),
        'company_id': fields.many2one('res.company', 'Company', readonly=True)
    }    
    _defaults={
        'period_id': _get_period,
        'target_move': 'posted',
        'result_selection': 'customer',
        'company_id': lambda s,cr,uid,c: s.pool.get('res.company')._company_default_get(cr, uid, 'account.account', context=c),
    }

    def print_report(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        data = self.read(cr, uid, ids)[0]       
        period_id = data['period_id'][0]
        company_id = data['company_id'][0]        
        data.update( {
            'ids': [period_id], 
            'period_id': period_id, 
            'company_id': company_id,        
            'model': 'account.period',
        } )
        return {
            'type': 'ir.actions.report.xml',
            'report_name': 'account.partner.open.arap.period.print',
            'datas': data,
        }

wiz_partner_open_arap_period()

